@@ -12,35 +12,34 @@ from gxformat2 import (
     convert_and_import_workflow,
     ImporterGalaxyInterface,
 )
-try:
-    from pyvirtualdisplay import Display
-except ImportError:
-    Display = None
-from six.moves.urllib.parse import urljoin
+from requests.models import Response
 
 from galaxy.selenium import (
     driver_factory,
 )
+from galaxy.selenium.context import (
+    GalaxySeleniumContext,
+)
 from galaxy.selenium.navigates_galaxy import (
-    NavigatesGalaxy,
     retry_during_transitions
 )
-from galaxy.util import asbool, classproperty
+from galaxy.util import (
+    asbool,
+    classproperty,
+    DEFAULT_SOCKET_TIMEOUT,
+)
 from galaxy_test.base import populators
 from galaxy_test.base.api import UsesApiTestCaseMixin
+from galaxy_test.base.api_util import get_admin_api_key
 from galaxy_test.base.env import DEFAULT_WEB_HOST, get_ip_address
 from galaxy_test.base.testcase import FunctionalTestCase
 try:
     from galaxy_test.driver.driver_util import GalaxyTestDriver
 except ImportError:
-    GalaxyTestDriver = None
+    GalaxyTestDriver = None  # type: ignore
 
 DEFAULT_TIMEOUT_MULTIPLIER = 1
 DEFAULT_TEST_ERRORS_DIRECTORY = os.path.abspath("database/test_errors")
-DEFAULT_SELENIUM_BROWSER = "auto"
-DEFAULT_SELENIUM_REMOTE = False
-DEFAULT_SELENIUM_REMOTE_PORT = "4444"
-DEFAULT_SELENIUM_REMOTE_HOST = "127.0.0.1"
 DEFAULT_SELENIUM_HEADLESS = "auto"
 DEFAULT_ADMIN_USER = "test@bx.psu.edu"
 DEFAULT_ADMIN_PASSWORD = "testpass"
@@ -51,10 +50,10 @@ TIMEOUT_MULTIPLIER = float(os.environ.get("GALAXY_TEST_TIMEOUT_MULTIPLIER", DEFA
 GALAXY_TEST_ERRORS_DIRECTORY = os.environ.get("GALAXY_TEST_ERRORS_DIRECTORY", DEFAULT_TEST_ERRORS_DIRECTORY)
 GALAXY_TEST_SCREENSHOTS_DIRECTORY = os.environ.get("GALAXY_TEST_SCREENSHOTS_DIRECTORY", None)
 # Test browser can be ["CHROME", "FIREFOX", "OPERA", "PHANTOMJS"]
-GALAXY_TEST_SELENIUM_BROWSER = os.environ.get("GALAXY_TEST_SELENIUM_BROWSER", DEFAULT_SELENIUM_BROWSER)
-GALAXY_TEST_SELENIUM_REMOTE = os.environ.get("GALAXY_TEST_SELENIUM_REMOTE", DEFAULT_SELENIUM_REMOTE)
-GALAXY_TEST_SELENIUM_REMOTE_PORT = os.environ.get("GALAXY_TEST_SELENIUM_REMOTE_PORT", DEFAULT_SELENIUM_REMOTE_PORT)
-GALAXY_TEST_SELENIUM_REMOTE_HOST = os.environ.get("GALAXY_TEST_SELENIUM_REMOTE_HOST", DEFAULT_SELENIUM_REMOTE_HOST)
+GALAXY_TEST_SELENIUM_BROWSER = os.environ.get("GALAXY_TEST_SELENIUM_BROWSER", driver_factory.DEFAULT_SELENIUM_BROWSER)
+GALAXY_TEST_SELENIUM_REMOTE = os.environ.get("GALAXY_TEST_SELENIUM_REMOTE", driver_factory.DEFAULT_SELENIUM_REMOTE)
+GALAXY_TEST_SELENIUM_REMOTE_PORT = os.environ.get("GALAXY_TEST_SELENIUM_REMOTE_PORT", driver_factory.DEFAULT_SELENIUM_REMOTE_PORT)
+GALAXY_TEST_SELENIUM_REMOTE_HOST = os.environ.get("GALAXY_TEST_SELENIUM_REMOTE_HOST", driver_factory.DEFAULT_SELENIUM_REMOTE_HOST)
 GALAXY_TEST_SELENIUM_HEADLESS = os.environ.get("GALAXY_TEST_SELENIUM_HEADLESS", DEFAULT_SELENIUM_HEADLESS)
 GALAXY_TEST_EXTERNAL_FROM_SELENIUM = os.environ.get("GALAXY_TEST_EXTERNAL_FROM_SELENIUM", None)
 # Auto-retry selenium tests this many times.
@@ -64,6 +63,7 @@ GALAXY_TEST_SELENIUM_USER_EMAIL = os.environ.get("GALAXY_TEST_SELENIUM_USER_EMAI
 GALAXY_TEST_SELENIUM_USER_PASSWORD = os.environ.get("GALAXY_TEST_SELENIUM_USER_PASSWORD", None)
 GALAXY_TEST_SELENIUM_ADMIN_USER_EMAIL = os.environ.get("GALAXY_TEST_SELENIUM_ADMIN_USER_EMAIL", DEFAULT_ADMIN_USER)
 GALAXY_TEST_SELENIUM_ADMIN_USER_PASSWORD = os.environ.get("GALAXY_TEST_SELENIUM_ADMIN_USER_PASSWORD", DEFAULT_ADMIN_PASSWORD)
+GALAXY_TEST_SELENIUM_BETA_HISTORY = os.environ.get("GALAXY_TEST_SELENIUM_BETA_HISTORY", "0") == "1"
 
 # JS code to execute in Galaxy JS console to setup localStorage of session for logging and
 # logging "flatten" messages because it seems Selenium (with Chrome at least) only grabs
@@ -99,7 +99,7 @@ def managed_history(f):
                 try:
                     current_history_id = self.current_history_id()
                     self.dataset_populator.cancel_history_jobs(current_history_id)
-                    self.api_delete("histories/%s" % current_history_id)
+                    self.api_delete(f"histories/{current_history_id}")
                 except Exception:
                     print("Faild to cleanup managed history, selenium connection corrupted somehow?")
     return func_wrapper
@@ -134,9 +134,9 @@ def dump_test_information(self, name_prefix):
         for log_type in ["browser", "driver"]:
             try:
                 full_log = self.driver.get_log(log_type)
-                trimmed_log = [l for l in full_log if l["level"] not in ["DEBUG", "INFO"]]
-                write_file("%s.log.json" % log_type, json.dumps(trimmed_log, indent=True))
-                write_file("%s.log.verbose.json" % log_type, json.dumps(full_log, indent=True))
+                trimmed_log = [entry for entry in full_log if entry["level"] not in ["DEBUG", "INFO"]]
+                write_file(f"{log_type}.log.json", json.dumps(trimmed_log, indent=True))
+                write_file(f"{log_type}.log.verbose.json", json.dumps(full_log, indent=True))
             except Exception:
                 continue
 
@@ -161,7 +161,7 @@ def selenium_test(f):
                 dump_test_information(self, test_name)
                 if retry_attempts < GALAXY_TEST_SELENIUM_RETRIES:
                     retry_attempts += 1
-                    print("Test function [{}] threw an exception, retrying. Failed attempts - {}.".format(test_name, retry_attempts))
+                    print(f"Test function [{test_name}] threw an exception, retrying. Failed attempts - {retry_attempts}.")
                 else:
                     raise
 
@@ -172,6 +172,7 @@ retry_assertion_during_transitions = partial(retry_during_transitions, exception
 
 
 class TestSnapshot:
+    __test__ = False  # Prevent pytest from discovering this class (issue #12071)
 
     def __init__(self, driver, index, description):
         self.screenshot_binary = driver.get_screenshot_as_png()
@@ -182,12 +183,31 @@ class TestSnapshot:
 
     def write_to_error_directory(self, write_file_func):
         prefix = "%d-%s" % (self.index, self.description)
-        write_file_func("%s-screenshot.png" % prefix, self.screenshot_binary, raw=True)
-        write_file_func("%s-traceback.txt" % prefix, self.exc)
-        write_file_func("%s-stack.txt" % prefix, str(self.stack))
+        write_file_func(f"{prefix}-screenshot.png", self.screenshot_binary, raw=True)
+        write_file_func(f"{prefix}-traceback.txt", self.exc)
+        write_file_func(f"{prefix}-stack.txt", str(self.stack))
 
 
-class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
+class GalaxyTestSeleniumContext(GalaxySeleniumContext):
+    """Extend GalaxySeleniumContext with Selenium-aware galaxy_test.base.populators."""
+
+    @property
+    def dataset_populator(self) -> populators.BaseDatasetPopulator:
+        """A dataset populator connected to the Galaxy session described by Selenium context."""
+        return SeleniumSessionDatasetPopulator(self)
+
+    @property
+    def dataset_collection_populator(self) -> populators.BaseDatasetCollectionPopulator:
+        """A dataset collection populator connected to the Galaxy session described by Selenium context."""
+        return SeleniumSessionDatasetCollectionPopulator(self)
+
+    @property
+    def workflow_populator(self) -> populators.BaseWorkflowPopulator:
+        """A workflow populator connected to the Galaxy session described by Selenium context."""
+        return SeleniumSessionWorkflowPopulator(self)
+
+
+class TestWithSeleniumMixin(GalaxyTestSeleniumContext, UsesApiTestCaseMixin):
     # If run one-off via nosetests, the next line ensures test
     # tools and datatypes are used instead of configured tools.
     framework_tool_and_types = True
@@ -198,6 +218,8 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
     # will be used to login.
     ensure_registered = False
 
+    ensure_beta_history = GALAXY_TEST_SELENIUM_BETA_HISTORY
+
     # Override this in subclasses to annotate that an admin user
     # is required for the test to run properly. Override admin user
     # login info with GALAXY_TEST_SELENIUM_ADMIN_USER_EMAIL /
@@ -226,7 +248,7 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
         try:
             self.setup_with_driver()
         except Exception:
-            dump_test_information(self, self.__class__.__name__ + "_setup")
+            dump_test_information(self, f"{self.__class__.__name__}_setup")
             raise
 
     def setup_with_driver(self):
@@ -237,6 +259,8 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
         """
         if self.ensure_registered:
             self.login()
+            if self.ensure_beta_history:
+                self.use_beta_history()
 
     def tear_down_selenium(self):
         self.tear_down_driver()
@@ -249,20 +273,6 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
         """
         self.snapshots.append(TestSnapshot(self.driver, len(self.snapshots), description))
 
-    def screenshot(self, label):
-        """If GALAXY_TEST_SCREENSHOTS_DIRECTORY is set create a screenshot there named <label>.png.
-
-        Unlike the above "snapshot" feature, this will be written out regardless and not in a per-test
-        directory. The above method is used for debugging failures within a specific test. This method
-        if more for creating a set of images to augment automated testing with manual human inspection
-        after a test or test suite has executed.
-        """
-        target = self._screenshot_path(label)
-        if target is None:
-            return
-
-        self.driver.save_screenshot(target)
-
     def get_download_path(self):
         """Returns default download path
         """
@@ -297,17 +307,13 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
 
     def reset_driver_and_session(self):
         self.tear_down_driver()
+        self.target_url_from_selenium = self._target_url_from_selenium()
         self.setup_driver_and_session()
         self._try_setup_with_driver()
 
     def setup_driver_and_session(self):
         self.display = driver_factory.virtual_display_if_enabled(use_virtual_display())
-        self.driver = get_driver()
-        # New workflow index page does not degrade well to smaller sizes, needed
-        # to increase this.
-        # Needed to up the height for paired list creator being taller in BS4 branch.
-        self.driver.set_window_size(1280, 1000)
-
+        self.configured_driver = get_configured_driver()
         self._setup_galaxy_logging()
 
     def _setup_galaxy_logging(self):
@@ -332,7 +338,7 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
             self.driver.close()
         except Exception as e:
             if "cannot kill Chrome" in str(e):
-                print("Ignoring likely harmless error in Selenium shutdown %s" % e)
+                print(f"Ignoring likely harmless error in Selenium shutdown {e}")
             else:
                 exception = e
 
@@ -352,13 +358,6 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
     def timeout_multiplier(self):
         return TIMEOUT_MULTIPLIER
 
-    def build_url(self, url, for_selenium=True):
-        if for_selenium:
-            base = self.target_url_from_selenium
-        else:
-            base = self.url
-        return urljoin(base, url)
-
     def assert_initial_history_panel_state_correct(self):
         # Move into a TestsHistoryPanel mixin
         unnamed_name = self.components.history_panel.new_name.text
@@ -370,7 +369,8 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
         initial_size_str = self.components.history_panel.new_size.text
         size_selector = self.components.history_panel.size
         size_text = size_selector.wait_for_text()
-        assert initial_size_str in size_text, "{} not in {}".format(initial_size_str, size_text)
+
+        assert initial_size_str in size_text, f"{initial_size_str} not in {size_text}"
 
         self.components.history_panel.empty_message.wait_for_visible()
 
@@ -382,18 +382,7 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
         )
         with self.main_panel():
             self.assert_no_error_message()
-
-    @property
-    def dataset_populator(self):
-        return SeleniumSessionDatasetPopulator(self)
-
-    @property
-    def dataset_collection_populator(self):
-        return SeleniumSessionDatasetCollectionPopulator(self)
-
-    @property
-    def workflow_populator(self):
-        return SeleniumSessionWorkflowPopulator(self)
+        return GALAXY_TEST_SELENIUM_ADMIN_USER_EMAIL
 
     def workflow_upload_yaml_with_random_name(self, content, **kwds):
         workflow_populator = self.workflow_populator
@@ -408,7 +397,7 @@ class TestWithSeleniumMixin(NavigatesGalaxy, UsesApiTestCaseMixin):
         """
         visualization_names = self.history_panel_item_available_visualizations(hid)
         if visualization_name not in visualization_names:
-            raise unittest.SkipTest("Skipping test, visualization [%s] doesn't appear to be configured." % visualization_name)
+            raise unittest.SkipTest(f"Skipping test, visualization [{visualization_name}] doesn't appear to be configured.")
 
 
 class SeleniumTestCase(FunctionalTestCase, TestWithSeleniumMixin):
@@ -417,6 +406,7 @@ class SeleniumTestCase(FunctionalTestCase, TestWithSeleniumMixin):
     def setUp(self):
         super().setUp()
         self.setup_selenium()
+        self.admin_api_key = get_admin_api_key()
 
     def tearDown(self):
         exception = None
@@ -465,6 +455,16 @@ class SharedStateSeleniumTestCase(SeleniumTestCase):
         """Override this to setup shared data for tests that gets initialized only once."""
 
 
+class UsesLibraryAssertions:
+
+    @retry_assertion_during_transitions
+    def assert_num_displayed_items_is(self, n):
+        self.assertEqual(n, self.num_displayed_items())
+
+    def num_displayed_items(self):
+        return len(self.libraries_table_elements())
+
+
 class UsesHistoryItemAssertions:
 
     def assert_item_peek_includes(self, hid, expected):
@@ -475,7 +475,7 @@ class UsesHistoryItemAssertions:
     def assert_item_info_includes(self, hid, expected):
         item_body = self.history_panel_item_component(hid=hid)
         info_text = item_body.info.wait_for_text()
-        assert expected in info_text, "Failed to find expected info text [{}] in info [{}]".format(expected, info_text)
+        assert expected in info_text, f"Failed to find expected info text [{expected}] in info [{info_text}]"
 
     def assert_item_dbkey_displayed_as(self, hid, dbkey):
         item_body = self.history_panel_item_component(hid=hid)
@@ -485,7 +485,7 @@ class UsesHistoryItemAssertions:
     def assert_item_summary_includes(self, hid, expected_text):
         item_body = self.history_panel_item_component(hid=hid)
         summary_text = item_body.summary.wait_for_text()
-        assert expected_text in summary_text, "Expected summary [{}] not found in [{}].".format(expected_text, summary_text)
+        assert expected_text in summary_text, f"Expected summary [{expected_text}] not found in [{summary_text}]."
 
     def assert_item_name(self, hid, expected_name):
         item_body = self.history_panel_item_component(hid=hid)
@@ -510,11 +510,14 @@ def default_web_host_for_selenium_tests():
         return DEFAULT_WEB_HOST
 
 
-def get_driver():
-    if asbool(GALAXY_TEST_SELENIUM_REMOTE):
-        return get_remote_driver()
-    else:
-        return get_local_driver()
+def get_configured_driver():
+    return driver_factory.ConfiguredDriver(
+        browser=GALAXY_TEST_SELENIUM_BROWSER,
+        remote=asbool(GALAXY_TEST_SELENIUM_REMOTE),
+        remote_host=GALAXY_TEST_SELENIUM_REMOTE_HOST,
+        remote_port=GALAXY_TEST_SELENIUM_REMOTE_PORT,
+        headless=headless_selenium(),
+    )
 
 
 def headless_selenium():
@@ -543,31 +546,27 @@ def use_virtual_display():
         return asbool(GALAXY_TEST_SELENIUM_HEADLESS)
 
 
-def get_local_driver():
-    return driver_factory.get_local_driver(
-        GALAXY_TEST_SELENIUM_BROWSER,
-        headless_selenium()
-    )
-
-
-def get_remote_driver():
-    return driver_factory.get_remote_driver(
-        host=GALAXY_TEST_SELENIUM_REMOTE_HOST,
-        port=GALAXY_TEST_SELENIUM_REMOTE_PORT,
-        browser=GALAXY_TEST_SELENIUM_BROWSER,
-    )
-
-
 class SeleniumSessionGetPostMixin:
     """Mixin for adapting Galaxy testing populators helpers to Selenium session backed bioblend."""
+    selenium_context: GalaxySeleniumContext
 
-    def _get(self, route, data={}):
-        full_url = self.selenium_test_case.build_url("api/" + route, for_selenium=False)
-        response = requests.get(full_url, data=data, cookies=self.selenium_test_case.selenium_to_requests_cookies())
+    @property
+    def _mixin_admin_api_key(self) -> str:
+        return getattr(self, "admin_api_key", get_admin_api_key())
+
+    def _get(self, route, data=None, headers=None, admin=False) -> Response:
+        data = data or {}
+        full_url = self.selenium_context.build_url(f"api/{route}", for_selenium=False)
+        cookies = None
+        if admin:
+            full_url = f"{full_url}?key={self._mixin_admin_api_key}"
+        else:
+            cookies = self.selenium_context.selenium_to_requests_cookies()
+        response = requests.get(full_url, params=data, cookies=cookies, headers=headers, timeout=DEFAULT_SOCKET_TIMEOUT)
         return response
 
-    def _post(self, route, data=None, files=None):
-        full_url = self.selenium_test_case.build_url("api/" + route, for_selenium=False)
+    def _post(self, route, data=None, files=None, headers=None, admin=False, json: bool = False) -> Response:
+        full_url = self.selenium_context.build_url(f"api/{route}", for_selenium=False)
         if data is None:
             data = {}
 
@@ -576,60 +575,80 @@ class SeleniumSessionGetPostMixin:
             if files is not None:
                 del data["__files"]
 
-        response = requests.post(full_url, data=data, cookies=self.selenium_test_case.selenium_to_requests_cookies(), files=files)
+        cookies = None
+        if admin:
+            full_url = f"{full_url}?key={self._mixin_admin_api_key}"
+        else:
+            cookies = self.selenium_context.selenium_to_requests_cookies()
+        response = requests.post(full_url, data=data, cookies=cookies, files=files, headers=headers, timeout=DEFAULT_SOCKET_TIMEOUT)
         return response
 
-    def _delete(self, route, data={}):
-        full_url = self.selenium_test_case.build_url("api/" + route, for_selenium=False)
-        response = requests.delete(full_url, data=data, cookies=self.selenium_test_case.selenium_to_requests_cookies())
+    def _delete(self, route, data=None, headers=None, admin=False, json: bool = False) -> Response:
+        data = data or {}
+        full_url = self.selenium_context.build_url(f"api/{route}", for_selenium=False)
+        cookies = None
+        if admin:
+            full_url = f"{full_url}?key={self._mixin_admin_api_key}"
+        else:
+            cookies = self.selenium_context.selenium_to_requests_cookies()
+        response = requests.delete(full_url, data=data, cookies=cookies, headers=headers, timeout=DEFAULT_SOCKET_TIMEOUT)
         return response
 
-    def __url(self, route):
-        return self._gi.url + "/" + route
+    def _put(self, route, data=None, headers=None, admin=False, json: bool = False) -> Response:
+        data = data or {}
+        full_url = self.selenium_context.build_url(f"api/{route}", for_selenium=False)
+        cookies = None
+        if admin:
+            full_url = f"{full_url}?key={self._mixin_admin_api_key}"
+        else:
+            cookies = self.selenium_context.selenium_to_requests_cookies()
+        response = requests.put(full_url, data=data, cookies=cookies, headers=headers, timeout=DEFAULT_SOCKET_TIMEOUT)
+        return response
 
 
-class SeleniumSessionDatasetPopulator(populators.BaseDatasetPopulator, SeleniumSessionGetPostMixin):
+class SeleniumSessionDatasetPopulator(SeleniumSessionGetPostMixin, populators.BaseDatasetPopulator):
 
     """Implementation of BaseDatasetPopulator backed by bioblend."""
 
-    def __init__(self, selenium_test_case):
+    def __init__(self, selenium_context: GalaxySeleniumContext):
         """Construct a dataset populator from a bioblend GalaxyInstance."""
-        self.selenium_test_case = selenium_test_case
+        self.selenium_context = selenium_context
 
 
-class SeleniumSessionDatasetCollectionPopulator(populators.BaseDatasetCollectionPopulator, SeleniumSessionGetPostMixin):
+class SeleniumSessionDatasetCollectionPopulator(SeleniumSessionGetPostMixin, populators.BaseDatasetCollectionPopulator):
 
     """Implementation of BaseDatasetCollectionPopulator backed by bioblend."""
 
-    def __init__(self, selenium_test_case):
+    def __init__(self, selenium_context: GalaxySeleniumContext):
         """Construct a dataset collection populator from a bioblend GalaxyInstance."""
-        self.selenium_test_case = selenium_test_case
-        self.dataset_populator = SeleniumSessionDatasetPopulator(selenium_test_case)
+        self.selenium_context = selenium_context
+        self.dataset_populator = SeleniumSessionDatasetPopulator(selenium_context)
 
-    def _create_collection(self, payload):
+    def _create_collection(self, payload: dict) -> Response:
         create_response = self._post("dataset_collections", data=payload)
         return create_response
 
 
-class SeleniumSessionWorkflowPopulator(populators.BaseWorkflowPopulator, SeleniumSessionGetPostMixin, ImporterGalaxyInterface):
+class SeleniumSessionWorkflowPopulator(SeleniumSessionGetPostMixin, populators.BaseWorkflowPopulator, ImporterGalaxyInterface):
 
     """Implementation of BaseWorkflowPopulator backed by bioblend."""
 
-    def __init__(self, selenium_test_case):
+    def __init__(self, selenium_context: GalaxySeleniumContext):
         """Construct a workflow populator from a bioblend GalaxyInstance."""
-        self.selenium_test_case = selenium_test_case
-        self.dataset_populator = SeleniumSessionDatasetPopulator(selenium_test_case)
+        self.selenium_context = selenium_context
+        self.dataset_populator = SeleniumSessionDatasetPopulator(selenium_context)
+        self.dataset_collection_populator = SeleniumSessionDatasetPopulator(selenium_context)
 
-    def import_workflow(self, workflow, **kwds):
+    def import_workflow(self, workflow: dict, **kwds) -> dict:
         workflow_str = json.dumps(workflow, indent=4)
         data = {
             'workflow': workflow_str,
         }
         data.update(**kwds)
         upload_response = self._post("workflows", data=data)
-        assert upload_response.status_code == 200
+        upload_response.raise_for_status()
         return upload_response.json()
 
-    def upload_yaml_workflow(self, has_yaml, **kwds):
+    def upload_yaml_workflow(self, has_yaml, **kwds) -> str:
         workflow = convert_and_import_workflow(has_yaml, galaxy_interface=self, **kwds)
         return workflow["id"]
