@@ -10,7 +10,6 @@ class HistoryOptionsTestCase(SeleniumTestCase):
     def test_options(self):
         self.register()
         self.perform_upload(self.get_filename("1.txt"))
-        self.wait_for_history()
         self.click_history_options()
 
         menu_selector = self.navigation.history_panel.selectors.options_menu
@@ -22,6 +21,7 @@ class HistoryOptionsTestCase(SeleniumTestCase):
         self.wait_for_absent_or_hidden(menu_selector)
 
         hid = 1
+        self.history_panel_wait_for_hid_state(hid, 'ok')
         self.history_panel_click_item_title(hid=hid, wait=True)
         item_component = self.history_panel_item_body_component(hid=hid)
         item_component.wait_for_visible()
