@@ -1,7 +1,12 @@
 import json
 
+import yaml
+from selenium.webdriver.common.keys import Keys
+
 from galaxy_test.base.workflow_fixtures import (
     WORKFLOW_NESTED_SIMPLE,
+    WORKFLOW_OPTIONAL_TRUE_INPUT_COLLECTION,
+    WORKFLOW_SELECT_FROM_OPTIONAL_DATASET,
     WORKFLOW_SIMPLE_CAT_TWICE,
     WORKFLOW_SIMPLE_MAPPING,
     WORKFLOW_WITH_INVALID_STATE,
@@ -27,16 +32,17 @@ class WorkflowEditorTestCase(SeleniumTestCase):
         editor = self.components.workflow_editor
         annotation = "basic_test"
         name = self.workflow_create_new(annotation=annotation)
-        edit_name_element = self.components.workflow_editor.edit_name.wait_for_visible()
-        actual_name = edit_name_element.get_attribute("value")
-        assert name in actual_name, "'{}' unequal name '{}'".format(name, actual_name)
-        edit_annotation_element = self.components.workflow_editor.edit_annotation.wait_for_visible()
-        actual_annotation = edit_annotation_element.get_attribute("value")
-        assert annotation in actual_annotation, "'{}' unequal annotation '{}'".format(annotation, actual_annotation)
+        self.assert_wf_name_is(name)
+        self.assert_wf_annotation_is(annotation)
 
         editor.canvas_body.wait_for_visible()
         editor.tool_menu.wait_for_visible()
 
+        # shouldn't have changes on fresh load
+        save_button = self.components.workflow_editor.save_button
+        save_button.wait_for_visible()
+        assert save_button.has_class("disabled")
+
         self.screenshot("workflow_editor_blank")
 
         self.components._.left_panel_drag.wait_for_visible()
@@ -53,6 +59,68 @@ class WorkflowEditorTestCase(SeleniumTestCase):
 
         self.screenshot("workflow_editor_left_and_right_collapsed")
 
+    @selenium_test
+    def test_edit_annotation(self):
+        editor = self.components.workflow_editor
+        annotation = "new_annotation_test"
+        name = self.workflow_create_new(annotation=annotation)
+        edit_annotation = self.components.workflow_editor.edit_annotation
+        self.assert_wf_annotation_is(annotation)
+
+        editor.canvas_body.wait_for_visible()
+
+        new_annotation = 'look new annotation'
+        edit_annotation.wait_for_and_send_keys(new_annotation)
+        self.assert_has_changes_and_save()
+        self.workflow_index_open_with_name(name)
+        self.assert_wf_annotation_is(new_annotation)
+
+    @selenium_test
+    def test_edit_name(self):
+        editor = self.components.workflow_editor
+        name = self.workflow_create_new()
+        editor.canvas_body.wait_for_visible()
+        new_name = self._get_random_name()
+        edit_name = self.components.workflow_editor.edit_name
+        edit_name.wait_for_and_send_keys(new_name)
+
+        self.assert_has_changes_and_save()
+        self.workflow_index_open_with_name(new_name)
+        self.assert_wf_name_is(name)
+
+    @selenium_test
+    def test_optional_select_data_field(self):
+        editor = self.components.workflow_editor
+        workflow_id = self.workflow_populator.upload_yaml_workflow(WORKFLOW_SELECT_FROM_OPTIONAL_DATASET)
+        self.workflow_index_open()
+        self.workflow_index_click_option("Edit")
+        editor = self.components.workflow_editor
+        node = editor.node._(label="select_from_dataset_optional")
+        node.title.wait_for_and_click()
+        self.components.tool_form.parameter_checkbox(parameter='select_single').wait_for_and_click()
+        # External (selenium-side) debounce hack for old backbone input
+        # TODO: remove when form elements are all converted.
+        self.components.tool_form.parameter_input(parameter='select_single').wait_for_and_send_keys('parameter valu')
+        self.sleep_for(self.wait_types.UX_RENDER)
+        self.components.tool_form.parameter_input(parameter='select_single').wait_for_and_send_keys('e')
+        self.sleep_for(self.wait_types.UX_RENDER)
+        self.assert_has_changes_and_save()
+        workflow = self.workflow_populator.download_workflow(workflow_id)
+        tool_state = json.loads(workflow['steps']['0']['tool_state'])
+        assert tool_state['select_single'] == 'parameter value'
+        # Disable optional button, resets value to null
+        self.components.tool_form.parameter_checkbox(parameter='select_single').wait_for_and_click()
+        self.assert_has_changes_and_save()
+        workflow = self.workflow_populator.download_workflow(workflow_id)
+        tool_state = json.loads(workflow['steps']['0']['tool_state'])
+        assert tool_state['select_single'] is None
+        # Enable button but don't provide a value
+        self.components.tool_form.parameter_checkbox(parameter='select_single').wait_for_and_click()
+        self.assert_has_changes_and_save()
+        workflow = self.workflow_populator.download_workflow(workflow_id)
+        tool_state = json.loads(workflow['steps']['0']['tool_state'])
+        assert tool_state['select_single'] == ""
+
     @selenium_test
     def test_data_input(self):
         editor = self.components.workflow_editor
@@ -63,9 +131,6 @@ class WorkflowEditorTestCase(SeleniumTestCase):
         editor.label_input.wait_for_and_send_keys("input1")
         editor.annotation_input.wait_for_and_send_keys("my cool annotation")
         self.sleep_for(self.wait_types.UX_RENDER)
-        self.screenshot("workflow_editor_data_input_filled_in PRECLICK")
-        editor.label_input.wait_for_and_click()  # Seems to help force the save of whole annotation.
-        self.sleep_for(self.wait_types.UX_RENDER)
         self.screenshot("workflow_editor_data_input_filled_in")
         self.workflow_editor_click_save()
 
@@ -75,7 +140,7 @@ class WorkflowEditorTestCase(SeleniumTestCase):
         label = editor.label_input.wait_for_value()
         assert label == "input1", label
         annotation = editor.annotation_input.wait_for_value()
-        assert annotation == "my cool annotation"
+        assert annotation == "my cool annotation", annotation
         data_input_node.destroy.wait_for_and_click()
         data_input_node.wait_for_absent()
         self.screenshot("workflow_editor_data_input_deleted")
@@ -89,7 +154,6 @@ class WorkflowEditorTestCase(SeleniumTestCase):
         self.screenshot("workflow_editor_data_collection_input_new")
         editor.label_input.wait_for_and_send_keys("input1")
         editor.annotation_input.wait_for_and_send_keys("my cool annotation")
-        editor.label_input.wait_for_and_click()  # Seems to help force the save of whole annotation.
         self.sleep_for(self.wait_types.UX_RENDER)
         self.screenshot("workflow_editor_data_collection_input_filled_in")
         self.workflow_editor_click_save()
@@ -100,12 +164,40 @@ class WorkflowEditorTestCase(SeleniumTestCase):
         label = editor.label_input.wait_for_value()
         assert label == "input1", label
         annotation = editor.annotation_input.wait_for_value()
-        assert annotation == "my cool annotation"
+        assert annotation == "my cool annotation", annotation
         data_input_node.destroy.wait_for_and_click()
         data_input_node.wait_for_absent()
         self.sleep_for(self.wait_types.UX_RENDER)
         self.screenshot("workflow_editor_data_collection_input_deleted")
 
+    @selenium_test
+    def test_data_column_input_editing(self):
+        self.open_in_workflow_editor("""
+class: GalaxyWorkflow
+steps:
+  column_param_list:
+    tool_id: column_param_list
+    state:
+      col: ["1","2","3"]
+      col_names: ["a", "b", "c"]
+      """)
+        editor = self.components.workflow_editor
+        node = editor.node._(label="column_param_list")
+        node.title.wait_for_and_click()
+        columns = self.components.tool_form.parameter_textarea(parameter='col')
+        textarea_columns = columns.wait_for_visible()
+        assert textarea_columns.get_attribute('value') == '1\n2\n3\n'
+        column_names = self.components.tool_form.parameter_textarea(parameter='col_names')
+        textarea_column_names = column_names.wait_for_visible()
+        assert textarea_column_names.get_attribute('value') == 'a\nb\nc\n'
+        self.set_text_element(columns, '4\n5\n6\n')
+        self.sleep_for(self.wait_types.UX_RENDER)
+        self.assert_has_changes_and_save()
+        self.driver.refresh()
+        node.title.wait_for_and_click()
+        textarea_columns = columns.wait_for_visible()
+        assert textarea_columns.get_attribute('value') == '4\n5\n6\n'
+
     @selenium_test
     def test_integer_input(self):
         editor = self.components.workflow_editor
@@ -115,7 +207,6 @@ class WorkflowEditorTestCase(SeleniumTestCase):
         self.screenshot("workflow_editor_parameter_input_new")
         editor.label_input.wait_for_and_send_keys("input1")
         editor.annotation_input.wait_for_and_send_keys("my cool annotation")
-        editor.label_input.wait_for_and_click()  # Seems to help force the save of whole annotation.
         self.sleep_for(self.wait_types.UX_RENDER)
         self.screenshot("workflow_editor_parameter_input_filled_in")
         self.workflow_editor_click_save()
@@ -126,7 +217,7 @@ class WorkflowEditorTestCase(SeleniumTestCase):
         label = editor.label_input.wait_for_value()
         assert label == "input1", label
         annotation = editor.annotation_input.wait_for_value()
-        assert annotation == "my cool annotation"
+        assert annotation == "my cool annotation", annotation
         data_input_node.destroy.wait_for_and_click()
         data_input_node.wait_for_absent()
         self.sleep_for(self.wait_types.UX_RENDER)
@@ -146,16 +237,16 @@ steps:
       inttest: input_int
 """)
         self.screenshot("workflow_editor_parameter_connection_simple")
-        self.assert_connected("input_int#output", "simple_constructs#inttest")
+        self.assert_connected("input_int#output", "tool_exec#inttest")
 
         editor = self.components.workflow_editor
 
-        tool_node = editor.node._(label="simple_constructs")
+        tool_node = editor.node._(label="tool_exec")
         tool_input = tool_node.input_terminal(name="inttest")
         tool_input.wait_for_and_click()
 
         editor.connector_destroy_callout.wait_for_and_click()
-        self.assert_not_connected("input_int#output", "simple_constructs#inttest")
+        self.assert_not_connected("input_int#output", "tool_exec#inttest")
         self.screenshot("workflow_editor_parameter_connection_destroyed")
 
         # When connected, cannot turn it into a RuntimeValue..
@@ -171,14 +262,14 @@ steps:
         # Also the connector should disappear
         tool_input.wait_for_absent_or_hidden()
 
-        # Now make it connected again and watch the requestss
+        # Now make it connected again and watch the requests
         connect_icon.wait_for_and_click()
 
         tool_input.wait_for_visible()
         collapse_input.wait_for_absent_or_hidden()
 
-        self.workflow_editor_connect("input_int#output", "simple_constructs#inttest", screenshot_partial="workflow_editor_parameter_connection_dragging")
-        self.assert_connected("input_int#output", "simple_constructs#inttest")
+        self.workflow_editor_connect("input_int#output", "tool_exec#inttest", screenshot_partial="workflow_editor_parameter_connection_dragging")
+        self.assert_connected("input_int#output", "tool_exec#inttest")
 
     @selenium_test
     def test_existing_connections(self):
@@ -200,6 +291,19 @@ steps:
         self.workflow_editor_connect("input1#output", "first_cat#input1", screenshot_partial="workflow_editor_connection_dragging")
         self.assert_connected("input1#output", "first_cat#input1")
 
+    @selenium_test
+    def test_reconnecting_nodes(self):
+        name = self.open_in_workflow_editor(WORKFLOW_SIMPLE_CAT_TWICE)
+        self.assert_connected("input1#output", "first_cat#input1")
+        self.workflow_editor_destroy_connection("first_cat#input1")
+        self.assert_not_connected("input1#output", "first_cat#input1")
+        self.workflow_editor_connect("input1#output", "first_cat#input1")
+        self.assert_connected("input1#output", "first_cat#input1")
+        self.sleep_for(self.wait_types.UX_RENDER)
+        self.assert_has_changes_and_save()
+        self.workflow_index_open_with_name(name)
+        self.assert_connected("input1#output", "first_cat#input1")
+
     @selenium_test
     def test_rendering_output_collection_connections(self):
         self.open_in_workflow_editor(WORKFLOW_WITH_OUTPUT_COLLECTION)
@@ -298,6 +402,30 @@ steps:
         self.screenshot("workflow_editor_edit_menu")
         self.workflow_editor_click_option("Save As")
 
+    @selenium_test
+    def test_editor_tool_upgrade(self):
+        workflow_populator = self.workflow_populator
+        workflow_id = workflow_populator.upload_yaml_workflow("""class: GalaxyWorkflow
+inputs: []
+steps:
+  - tool_id: multiple_versions
+    tool_version: 0.1
+    label: multiple_versions
+    state:
+      foo: bar
+        """, exact_tools=True)
+        self.workflow_index_open()
+        self.workflow_index_click_option("Edit")
+        editor = self.components.workflow_editor
+        editor.node._(label="multiple_versions").wait_for_and_click()
+        editor.tool_version_button.wait_for_and_click()
+        assert self.select_dropdown_item('Switch to 0.2'), 'Switch to tool version dropdown item not found'
+        self.screenshot("workflow_editor_version_update")
+        self.sleep_for(self.wait_types.UX_RENDER)
+        self.assert_has_changes_and_save()
+        workflow = self.workflow_populator.download_workflow(workflow_id)
+        assert workflow['steps']['0']['tool_version'] == '0.2'
+
     @selenium_test
     def test_editor_tool_upgrade_message(self):
         workflow_populator = self.workflow_populator
@@ -306,6 +434,109 @@ steps:
         self.workflow_index_click_option("Edit")
         self.assert_modal_has_text("Using version '0.2' instead of version '0.0.1'")
         self.screenshot("workflow_editor_tool_upgrade")
+        self.components.workflow_editor.modal_button_continue.wait_for_and_click()
+        self.assert_has_changes_and_save()
+
+    @selenium_test
+    def test_editor_subworkflow_tool_upgrade_message(self):
+        workflow_populator = self.workflow_populator
+        embedded_workflow = yaml.safe_load(WORKFLOW_WITH_OLD_TOOL_VERSION)
+        # Create invalid tool state
+        embedded_workflow['steps']['mul_versions']['state']['inttest'] = 'Invalid'
+        outer_workflow = yaml.safe_load("""
+class: GalaxyWorkflow
+inputs:
+  outer_input: data
+steps:
+  nested_workflow:
+    run: {}
+    in:
+      input1: outer_input
+        """)
+        outer_workflow['steps']['nested_workflow']['run'] = embedded_workflow
+        workflow_populator.upload_yaml_workflow(json.dumps(outer_workflow), exact_tools=True)
+        self.workflow_index_open()
+        self.workflow_index_click_option("Edit")
+        self.sleep_for(self.wait_types.UX_RENDER)
+        self.assert_modal_has_text("Using version '0.2' instead of version '0.0.1'")
+        self.assert_modal_has_text("parameter 'inttest': an integer or workflow parameter is required")
+        self.screenshot("workflow_editor_subworkflow_tool_upgrade")
+        self.components.workflow_editor.modal_button_continue.wait_for_and_click()
+        self.assert_has_changes_and_save()
+
+    @staticmethod
+    def set_text_element(element, value):
+        # Try both, no harm here
+        element.wait_for_and_send_keys(Keys.CONTROL, 'a')
+        element.wait_for_and_send_keys(Keys.COMMAND, 'a')
+        element.wait_for_and_send_keys(Keys.BACKSPACE)
+        element.wait_for_and_send_keys(value)
+
+    @selenium_test
+    def test_editor_duplicate_node(self):
+        workflow_id = self.workflow_populator.upload_yaml_workflow(WORKFLOW_SIMPLE_CAT_TWICE)
+        self.workflow_index_open()
+        self.workflow_index_click_option("Edit")
+        editor = self.components.workflow_editor
+        cat_node = editor.node._(label="first_cat")
+        cat_node.wait_for_and_click()
+        self.set_text_element(editor.label_input, 'source label')
+        # Select node using new label, ensures labels are synced between side panel and node
+        cat_node = editor.node._(label="source label")
+        self.assert_has_changes_and_save()
+        editor.annotation_input.wait_for_and_send_keys("source annotation")
+        self.assert_has_changes_and_save()
+        editor.configure_output(output='out_file1').wait_for_and_click()
+        output_label = editor.label_output(output='out_file1')
+        self.set_text_element(output_label, 'workflow output label')
+        self.set_text_element(editor.rename_output, 'renamed_output')
+        editor.change_datatype.wait_for_and_click()
+        editor.select_datatype(datatype='bam').wait_for_and_click()
+        self.set_text_element(editor.add_tags, '#crazynewtag')
+        self.set_text_element(editor.remove_tags, '#oldboringtag')
+        self.sleep_for(self.wait_types.UX_RENDER)
+        cat_node.clone.wait_for_and_click()
+        editor.label_input.wait_for_and_send_keys('cloned label')
+        output_label = editor.label_output(output='out_file1')
+        self.set_text_element(output_label, 'cloned output label')
+        self.assert_has_changes_and_save()
+        edited_workflow = self.workflow_populator.download_workflow(workflow_id)
+        source_step = next(iter(step for step in edited_workflow['steps'].values() if step['label'] == 'source label'))
+        cloned_step = next(iter(step for step in edited_workflow['steps'].values() if step['label'] == 'cloned label'))
+        assert source_step['annotation'] == cloned_step['annotation'] == "source annotation"
+        assert source_step['workflow_outputs'][0]['label'] == 'workflow output label'
+        assert cloned_step['workflow_outputs'][0]['label'] == 'cloned output label'
+        assert len(source_step['post_job_actions']) == len(cloned_step['post_job_actions']) == 4
+        assert source_step['post_job_actions'] == cloned_step['post_job_actions']
+
+    @selenium_test
+    def test_editor_embed_workflow(self):
+        workflow_populator = self.workflow_populator
+        child_workflow_name = self._get_random_name()
+        workflow_populator.upload_yaml_workflow(WORKFLOW_OPTIONAL_TRUE_INPUT_COLLECTION, name=child_workflow_name)
+        parent_workflow_id = workflow_populator.upload_yaml_workflow("""class: GalaxyWorkflow
+inputs: []
+steps:
+  - tool_id: multiple_versions
+    tool_version: 0.1
+    label: multiple_versions
+    state:
+      foo: bar
+        """)
+        self.workflow_index_open()
+        self.workflow_index_click_option("Edit")
+        editor = self.components.workflow_editor
+        editor.canvas_body.wait_for_visible()
+        editor.tool_menu.wait_for_visible()
+        editor.tool_menu_section_link(section_name="workflows").wait_for_and_click()
+        editor.workflow_link(workflow_title=child_workflow_name).wait_for_and_click()
+        self.sleep_for(self.wait_types.UX_RENDER)
+        self.assert_has_changes_and_save()
+        workflow = self.workflow_populator.download_workflow(parent_workflow_id)
+        subworkflow_step = workflow['steps']['1']
+        assert subworkflow_step['name'] == child_workflow_name
+        assert subworkflow_step['type'] == 'subworkflow'
+        assert subworkflow_step['subworkflow']['a_galaxy_workflow'] == 'true'
 
     @selenium_test
     def test_editor_invalid_tool_state(self):
@@ -348,16 +579,9 @@ steps:
         # Assert workflow not initially bookmarked.
         assert_workflow_bookmarked_status(False)
 
-        save_button = self.components.workflows.save_button
-        save_button.wait_for_clickable()
-
-        # element is clickable, but still might be behind the modal
+        self.components.workflow_editor.canvas_body.wait_for_visible()
         self.wait_for_selector_absent_or_hidden(self.modal_body_selector())
-        save_button.wait_for_and_click()
-
-        # wait for saving
-        self.wait_for_selector_absent_or_hidden(self.modal_body_selector())
-        self.driver.find_element_by_id("workflow").click()
+        self.components.masthead.workflow.wait_for_and_click()
 
         # parse workflow table
         table_elements = self.workflow_index_table_elements()
@@ -382,8 +606,8 @@ steps:
 
     def workflow_editor_connect(self, source, sink, screenshot_partial=None):
         source_id, sink_id = self.workflow_editor_source_sink_terminal_ids(source, sink)
-        source_element = self.driver.find_element_by_css_selector("#" + source_id)
-        sink_element = self.driver.find_element_by_css_selector("#" + sink_id)
+        source_element = self.driver.find_element_by_css_selector(f"#{source_id}")
+        sink_element = self.driver.find_element_by_css_selector(f"#{sink_id}")
 
         ac = self.action_chains()
         ac = ac.move_to_element(source_element).click_and_hold()
@@ -409,6 +633,7 @@ steps:
         self.workflow_index_open()
         self.workflow_index_open_with_name(name)
         self.workflow_editor_click_option("Auto Layout")
+        return name
 
     def workflow_editor_source_sink_terminal_ids(self, source, sink):
         editor = self.components.workflow_editor
@@ -477,24 +702,29 @@ steps:
         workflow_populator.upload_yaml_workflow(content, name=name)
         return name
 
-    def workflow_create_new(self, annotation=None, clear_placeholder=False):
-        self.workflow_index_open()
+    @retry_assertion_during_transitions
+    def assert_has_changes_and_save(self):
+        save_button = self.components.workflow_editor.save_button
+        save_button.wait_for_visible()
+        assert not save_button.has_class("disabled")
+        save_button.wait_for_and_click()
         self.sleep_for(self.wait_types.UX_RENDER)
-        self.click_button_new_workflow()
-        self.sleep_for(self.wait_types.UX_RENDER)
-        form_element = self.driver.find_element_by_id("submit")
-        name = self._get_random_name()
-        annotation = annotation or self._get_random_name()
-        inputs = self.driver.find_elements_by_class_name("ui-input")
-        if clear_placeholder:
-            inputs[0].clear()
-        inputs[0].send_keys(name)
-        inputs[1].send_keys(annotation)
-        form_element.click()
-        return name
+
+    @retry_assertion_during_transitions
+    def assert_wf_name_is(self, expected_name):
+        edit_name_element = self.components.workflow_editor.edit_name.wait_for_visible()
+        actual_name = edit_name_element.get_attribute("value")
+        assert expected_name in actual_name, f"'{expected_name}' unequal name '{actual_name}'"
+
+    @retry_assertion_during_transitions
+    def assert_wf_annotation_is(self, expected_annotation):
+        edit_annotation = self.components.workflow_editor.edit_annotation
+        edit_annotation_element = edit_annotation.wait_for_visible()
+        actual_annotation = edit_annotation_element.get_attribute("value")
+        assert expected_annotation in actual_annotation, f"'{expected_annotation}' unequal annotation '{actual_annotation}'"
 
     @retry_assertion_during_transitions
     def assert_modal_has_text(self, expected_text):
-        modal_element = self.wait_for_selector_visible(self.modal_body_selector())
+        modal_element = self.components.workflow_editor.state_modal_body.wait_for_visible()
         text = modal_element.text
-        assert expected_text in text, "Failed to find expected text [{}] in modal text [{}]".format(expected_text, text)
+        assert expected_text in text, f"Failed to find expected text [{expected_text}] in modal text [{text}]"
